/*
 *  linux-identd - an identification daemon for Linux
 * 
 *  Copyright (c) 2001-2002 by Per Liden <per@fukt.bth.se>
 * 
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, 
 *  USA.
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include <signal.h>
#include <sysexits.h>
#include <syslog.h>
#include <pwd.h>
#include <libgen.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <sys/types.h>
#include <netinet/in.h>
#include <arpa/inet.h>

#define PORT                    113
#define BACKLOG                 10
#define TIMEOUT                 60
#define USER                    "nobody"

#define PROC_FILE               "/proc/net/tcp"
#define PROC_FILE_MAXLINE       256

#define TCP_INVALID_PORT        0
#define TCP_MIN_PORT            1
#define TCP_MAX_PORT            65535

#define LOOKUP_OK               0
#define LOOKUP_NOT_FOUND        1
#define LOOKUP_ERROR            2

#define MAX_RESPONSE            200
#define MAX_REQUEST             100

/* glibc 2.1.x doesn't declare in_addr_t */
#if defined(__GLIBC_MINOR__) && (__GLIBC_MINOR__ < 2)
typedef uint32_t in_addr_t;
#endif

static void session_terminate(int s)
{
   while(wait(0) > 0);
}

static void session_timeout(int sig)
{
   exit(EX_OK);
}

static int read_request(int fd, char* buffer, int maxlen)
{
   int bytesread = 0;
   char c;

   while (bytesread < maxlen - 1) {
      if (read(fd, &c, 1) != 1)
         return 0;
   
      if (c == '\n')
         break;
      
      buffer[bytesread++] = c;
   }
   
   if (bytesread > 0 && buffer[bytesread - 1] == '\r')
      bytesread--;

   buffer[bytesread] = '\0';

   return 1;
}

static int write_response(int fd, char* buffer, int len)
{
   int retval;
   int byteswritten = 0;

   while (byteswritten < len) {
      retval = write(fd, buffer + byteswritten, len - byteswritten);
      if (retval <= 0)
         return 0;

      byteswritten += retval;
   }
   
   return 1;
}

static int lookup_uid(in_addr_t l_addr, unsigned int l_port,
                      in_addr_t r_addr, unsigned int r_port, uid_t* uid)
{
   FILE* file;

   if ((file = fopen(PROC_FILE, "r")) == 0) {
      syslog(LOG_ERR, "Failed to open %s: %s", PROC_FILE, strerror(errno));
      return LOOKUP_ERROR;
   }

#ifdef DEBUG
   printf("Looking for: %.8X:%.4X %.8X:%.4X\n", l_addr, l_port, r_addr, r_port);
#endif

   while (!feof(file)) {      
      char buffer[PROC_FILE_MAXLINE];
      unsigned int la, ra, lp, rp;
      la = ra = lp = rp = 0;

      fgets(buffer, sizeof(buffer) - 1, file);
      sscanf(buffer, "%*d: %X:%X %X:%X %*X %*X:%*X %*X:%*X %*X %d", &la, &lp, &ra, &rp, uid);
      
      if (la == l_addr && ra == r_addr && lp == l_port && rp == r_port) {
#ifdef DEBUG
         printf("Found, uid = %d\n", *uid);
#endif
         fclose(file);
         return LOOKUP_OK;
      }
   }

#ifdef DEBUG
   printf("Not found\n");
#endif

   fclose(file);
   return LOOKUP_NOT_FOUND;
}

static void compose_response(const int in, const char* request, char* response, unsigned int* response_len)
{
   struct sockaddr_in l_addr, r_addr;
   unsigned int l_port, r_port;
   struct passwd* pwd;
   socklen_t addr_len;
   uid_t uid;

   addr_len = sizeof(struct sockaddr_in);
   
   if (getpeername(in, (struct sockaddr*)&r_addr, &addr_len) ||
       getsockname(in, (struct sockaddr*)&l_addr, &addr_len)) {
      *response_len = snprintf(response, *response_len, "%.20s : ERROR : UNKNOWN-ERROR\r\n", request);
   } else {
      l_port = r_port = TCP_INVALID_PORT;
      sscanf(request, " %u , %u ", &l_port, &r_port);
   
      if (l_port < TCP_MIN_PORT || l_port > TCP_MAX_PORT || r_port < TCP_MIN_PORT || r_port > TCP_MAX_PORT) {
         *response_len = snprintf(response, *response_len, "%.20s : ERROR : INVALID-PORT\r\n", request);
      } else {
         int result = lookup_uid(l_addr.sin_addr.s_addr, l_port, r_addr.sin_addr.s_addr, r_port, &uid);
         
         if (result == LOOKUP_NOT_FOUND || (result == LOOKUP_OK && (pwd = getpwuid(uid)) == 0)) {
            *response_len = snprintf(response, *response_len, "%.20s : ERROR : NO-USER\r\n", request);
         } else if (result == LOOKUP_OK) {
            *response_len = snprintf(response, *response_len, "%.20s : USERID : UNIX : %.20s\r\n", request, pwd->pw_name);
         } else {
            *response_len = snprintf(response, *response_len, "%.20s : ERROR : UNKNOWN-ERROR\r\n", request);
         }
      }
   }
}

static void session_handler(int in, int out)
{
   unsigned int response_len;
   char response[MAX_RESPONSE];
   char request[MAX_REQUEST];
   
   signal(SIGALRM, session_timeout);
   alarm(TIMEOUT);
   
   for (;;) {
      if (!read_request(in, request, MAX_REQUEST))
         break;
      
#ifdef DEBUG
      printf("Request: %s\n", request);
#endif

      response_len = sizeof(response);
      compose_response(in, request, response, &response_len);

#ifdef DEBUG
      printf("Response: %s\n", response);
#endif

      if(!write_response(out, response, response_len))
         break;

      alarm(TIMEOUT);
   }
}

static void daemon_mode()
{
   int server_socket, session_socket;
   struct sockaddr_in l_addr, r_addr;
   struct passwd* pwd;
   int yes = 1;
   pid_t pid;

#ifndef DEBUG
   if (daemon(0, 0) == -1) {
      syslog(LOG_ERR, "Failed to become daemon: %s", strerror(errno));
      exit(EX_OSERR);
   }
#endif

   if ((server_socket = socket(AF_INET, SOCK_STREAM, 0)) == -1) {
      syslog(LOG_ERR, "Failed to create socket: %s", strerror(errno));
      exit(EX_OSERR);
   }

   if (setsockopt(server_socket, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) == -1) {
      syslog(LOG_ERR, "Failed to set socket options: %s", strerror(errno));
      exit(EX_OSERR);
   }
         
   memset(&l_addr, 0, sizeof(struct sockaddr_in));
   l_addr.sin_family = AF_INET;
   l_addr.sin_port = htons(PORT);
   l_addr.sin_addr.s_addr = INADDR_ANY;

   if (bind(server_socket, (struct sockaddr*)&l_addr, sizeof(struct sockaddr)) == -1) {
      syslog(LOG_ERR, "Failed to bind to port %d: %s", PORT, strerror(errno));
      exit(EX_OSERR);
   }

   if (listen(server_socket, BACKLOG) == -1) {
      syslog(LOG_ERR, "Failed to listen to port %d: %s", PORT, strerror(errno));
      exit(EX_OSERR);
   }

   if ((pwd = getpwnam(USER)) == 0) {
      syslog(LOG_ERR, "Failed to get user id of user %s, session terminated", USER);
      exit(EX_NOUSER);
   }
      
   if (setuid(pwd->pw_uid) == -1) {
      syslog(LOG_ERR, "Failed to set user id to %d (%s): %s", pwd->pw_uid, USER, strerror(errno));
      exit(EX_NOPERM);
   }

   signal(SIGCHLD, session_terminate);

   for(;;) {
      socklen_t addr_len = sizeof(struct sockaddr_in);
      if ((session_socket = accept(server_socket, (struct sockaddr*)&r_addr, &addr_len)) == -1) {
         syslog(LOG_ERR, "Failed to accept connection on port %d: %s", PORT, strerror(errno));
         continue;
      }

#ifdef DEBUG
      printf("Got connection from %s\n", inet_ntoa(r_addr.sin_addr));
#endif
      
      pid = fork();
      if (pid == 0) {
         close(server_socket);
         session_handler(session_socket, session_socket);
         close(session_socket);
         exit(EX_OK);
      } else if (pid == -1) {
         syslog(LOG_ERR, "Failed to create new process: %s", strerror(errno));
      }

      close(session_socket);
   }
}

static void inetd_mode()
{
   session_handler(fileno(stdin), fileno(stdout));
}

static void print_help()
{
   printf("identd version %s, Copyright (c) 2001-2002 by Per Liden <per@fukt.bth.se>\n", VERSION);
   printf("options:\n");
   printf("   -d    run in daemon mode\n");
   printf("   -v    print version and exit\n");
   printf("   -h    show this messagen\n");
#if 0
   /* Future features? */
   printf("   -o         do not reveal host type\n");
   printf("   -t <sec>   connection timeout (default is 60)\n");
   printf("   -u <user>  run as <user> (default is nobody)\n");
#endif
}

static void print_version()
{
   printf("identd %s\n", VERSION);
}

int main(int argc, char** argv)
{
   int i, become_daemon = 0;

   if (argc > 1) {
      for (i = 1; i < argc; i++) {
         if (!strcmp(argv[i], "-d")) {
            become_daemon = 1;
         } else if (!strcmp(argv[i], "-v")) {
            print_version();
            exit(EX_OK);
         } else if (!strcmp(argv[i], "-h")) {
            print_help();
            exit(EX_OK);
         } else {
            fprintf(stderr, "invalid option '%s'\n\n", argv[i]);
            print_help();
            exit(EX_USAGE);
         }
      }
   }

   openlog(0, LOG_PID, LOG_DAEMON);

   if (become_daemon)
      daemon_mode();
   else
      inetd_mode();

   return EX_OK;
} 
